
/*
 * Copyright (c) 2020, Alebo Industries, all rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the license along with this
 * program (LICENSE_GPL30.TXT). If not, see <https://www.gnu.org/licenses/>.
 */

/*
 * Please note that the GNU General Public License as specified above only
 * applies to the software in this file. If you have received the initial
 * TopDesign.cysch file as provided by Alebo Industries that specifies
 * hardware and component configurations configurations ("configuration
 * data") for Cypress' PSoC Creator software to implement an exemplary
 * instance of a design with which to test this program, insofar as the
 * configuration data as specified by Alebo Industries and captured in the
 * TopDesign.cysch file is not subject to further copyright or license
 * restrictions by Cypress Semiconductor Corp., the provided configuration
 * data only are distributed under the following permissive license (a
 * modified Zero-Clause BSD style license). The sine_1kHz.dat file, if
 * provided, is considered "configuration data" for licensing purposes
 * and is also distributed under this license.
 *
 *     Copyright (c) 2020 by Alebo Industries
 *
 *     Permission to use, copy, modify, and/or distribute this configuration
 *     data for any purpose with or without fee is hereby granted.
 *
 *     THE CONFIGURATION DATA IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS
 *     ALL WARRANTIES WITH REGARD TO THIS CONFIGURATION DATA INCLUDING ALL
 *     IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
 *     THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 *     DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 *     PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 *     ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 *     THIS CONFIGURATION DATA.
 *
 * Please consult any relevant licensing terms with regards to all other files
 * that may have been distributed with this program.
 */

#include <project.h>
#include <stdint.h>

/*
 * This project was develooped on a PSoC 4200 MCU, but could easily be adapted
 * to other platforms. To use it, make sure the HFClk is set to 48MHz. The PWM
 * cycle frequency is 22.050kHz, but each 11-bit audio sample is sent twice to
 * achieve an 11.025kHz sample rate. An audio frequency filter is required on
 * the PWM output. I used a 4th order Chebychev low pass filter with the two
 * on-chip op-amps in a Sallen-Key configuration with a 1kHz cutoff frequency
 * using resistors and capacitors. Note: The Cload specification for the on-chip
 * op-amps states that a maximum capacitive load of 125pF on these components.
 */

// Number of bits per audio sample
#define AUDIO_BIT_DEPTH 11

// This loads the audio data into the C code as a byte array
const uint8_t audio_data [] = {
#include "sine_1kHz.dat"  // Put comma separated values audio samples file here
};

// Use modulo 8 of sample_idx * 11 to mask off appropriate number of upper bits
const uint8_t mask_upper_bits [] = {
    0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03, 0x01
};

// Calculates the number of audio samples in the included data file
const size_t num_samples = sizeof(audio_data) * 8 / AUDIO_BIT_DEPTH;

// Added to sample_value before writing it [this value is calculated in main()]
uint16_t sample_offset = 0;

uint32_t fetch_next_sample(void)
{
    static size_t next_sample_idx = 0; // Used with bits_remaining to point at next bits to send
    static uint8_t flip_flop = 0;  // Used to send the same sample twice (22.050kHz PWM -> 11.025kHz audio)
    
    register uint32_t sample_value;
    register uint32_t bits_remaining;
    
    // Load masked bits from first byte into proper position
    bits_remaining = (AUDIO_BIT_DEPTH - 8) + next_sample_idx * AUDIO_BIT_DEPTH % 8;
    sample_value = (audio_data[next_sample_idx]
        & mask_upper_bits[next_sample_idx * AUDIO_BIT_DEPTH % 8]) << bits_remaining;
        
    if(bits_remaining > 8) // If the sample spans more than 2 bytes in the audio buffer
    {
        sample_value |= audio_data[next_sample_idx + 1] << (bits_remaining - 8);
        bits_remaining -= 8;
        if(bits_remaining > 0)
        {
            sample_value |= audio_data[next_sample_idx + 2] >> (8 - bits_remaining);
        }
    }   
    else
    {
         sample_value |= audio_data[next_sample_idx + 1] >> (8 - bits_remaining);
    }
    
    // Send each sample twice to get proper 11.025kHz sample rate with 22.05kHz PWM cycle frequency
    if(flip_flop == 0)
    {
        flip_flop = 1;  // Send the same sample next time as well
    }
    else
    {
        flip_flop = 0;
        next_sample_idx++;  // Send new sample next time
        next_sample_idx %= num_samples;  // If we've reached num_samples, wrap to zero
    }
    
    return(sample_value + sample_offset);
}

CY_ISR(Audio_PWM_Load_Sample)
{
    Audio_PWM_WriteCompareBuf(fetch_next_sample());
    Audio_PWM_ClearInterrupt(Audio_PWM_INTR_MASK_TC);
}

int main()
{
    Audio_PWM_TC_ISR_StartEx(Audio_PWM_Load_Sample);
    
    Audio_PWM_Init();  // So the Period value and other settings are loaded into PWM
    
    // Do once at start to centre samples in PWM range (is added to each sample before sending to PWM)
    sample_offset = (Audio_PWM_ReadPeriod() - ((uint32_t)0x01 << AUDIO_BIT_DEPTH)) / 2;
    
    // Write the first two samples into the PWM
    Audio_PWM_WriteCompare(fetch_next_sample());
    Audio_PWM_WriteCompareBuf(fetch_next_sample());
    
    CyGlobalIntEnable; // Enable global interrupts

    Audio_PWM_Enable();  // Only enables, does not re-initialize
    
    for(;;)
    {
        CySysPmSleep();
    }
}




